module FeatureStructure where

import Data.Set

data Element = Element { elName :: String, 
                         elType :: String, 
                         value :: String }
type Reference = (Element, Element) --FIXME: do not use Sets of References (Eq Element!)
type ReferenceExpl = ((String,Element), Element)
data Contract = Contract { contractName :: String,
                           contract :: Set Element -> Bool }
data Feature = Feature { featName :: String, --separated with '*' if non-atomic
                         elements :: Set Element,
                         references :: Set Reference,
                         contracts :: Set Contract }
type FeatureSet = Set Feature

instance Show Element where
    show (Element n t v) = n
instance Eq Element where
    (Element n1 t1 _) == (Element n2 t2 _) = n1==n2 && t1==t2
instance Ord Element where
    (Element n1 _ _) <= (Element n2 _ _) = n1 <= n2

instance Show Contract where
    show (Contract n c) = n
instance Eq Contract where
    (Contract n1 _) == (Contract n2 _) = n1 == n2
instance Ord Contract where
    (Contract n1 _) <= (Contract n2 _) = n1 <= n2

instance Show Feature where
    show (Feature n es rs cs) = n
instance Eq Feature where
    (Feature n1 _ _ _) == (Feature n2 _ _ _) = n1 == n2
instance Ord Feature where
    (Feature n1 _ _ _) <= (Feature n2 _ _ _) = n1 <= n2


-- search for features with given name
-- Note that this might be more difficult for composite features
getFeatureByName :: FeatureSet -> String -> Feature
getFeatureByName fs s = Data.Set.findMin 
    (Data.Set.filter (\f -> s == (featName f)) fs)

-- search for features with given element by element name
getFeaturesWithElement :: FeatureSet -> Element -> FeatureSet
getFeaturesWithElement fs e = Data.Set.filter (\f -> member e (elements f)) fs

-- search for features with exactly the given element by name and value
-- first get features with element by element name and then filter for
-- additional similarities
-- Usage: Dependencies need 
-- Running time: for a low numbers of features implementing e (#2) like getFeaturesWithElement
getFeaturesWithExactElement :: FeatureSet -> Element -> FeatureSet
getFeaturesWithExactElement fs e = Data.Set.filter 
    (\f -> or [True |  e2 <- (toList (Data.Set.filter (== e)  (elements f))),
                       elType e == elType e2, 
                       value e == value e2]) 
    (getFeaturesWithElement fs e)

-- checks if the given Featue combination has all containing reference targets
-- #1: features of combination with elements and references (base feature first)
-- running time: O( r * log(e) )  [r=#references, e=#elements]
isReferentiallyDependent :: [Feature] -> Bool
isReferentiallyDependent [] = False
isReferentiallyDependent fs = refDepHelp 
    (Data.Set.fold (Data.Set.union) Data.Set.empty (Data.Set.map (elements) (Data.Set.fromList fs))) 
    (Data.Set.toList (Data.Set.fold (Data.Set.union) Data.Set.empty (Data.Set.map (references) (Data.Set.fromList fs))))
    where refDepHelp :: Set Element -> [Reference] -> Bool
          refDepHelp es (r:rs) = (refDep1 es r) && (refDepHelp es rs)
          refDep1 :: Set Element -> Reference -> Bool
          refDep1 es r = (not (member (fst r) es)) || (member (snd r) es)

-- superimposition for elements: will not typecheck / check for soundness
-- composite element trees are merged - at present both trees' root names need to be equal
-- momentarily, same elements are overwritten
-- #1: feature to be composed with
-- #2: basis of composition
--superimposeElems :: Feature -> Feature -> Set Element
--superimposeElems f1 f2 = Data.Set.union (elements f2) (elements f1)


-- TODO not yet included
-- - type checks (like assignment compatibility)
-- - superimposition
