{
--
-- Parser for extracting references out of *.rsf files
--
-- Author: Wolfgang Scholz, 2010-06
--
module RSF where
import Char
import FeatureStructure
import Data.List


}

%name parseRSF
-- [Token] -> [Reference]
%tokentype { Token }
%error { parseError }

%token
    MEMBER          { TokenMember }
    CONTAINEDIN     { TokenContainedIn }
    COMPOUND        { TokenCompound }
    BASEDON         { TokenBasedOn }
    LOCATEDAT       { TokenLocatedAt }
    REFERSTO        { TokenRefersTo }
    REFvariable     { TokenRefVar }
    REFfunction     { TokenRefFunction }
    refClass        { TokenRefClass }
    refFile         { TokenRefFile }
    inheritClass    { TokenInheritClass }
    implements      { TokenImplements }
    wrappee         { TokenWrappee }
    var             { TokenElem $$ }

--%left '('
--%left not
%%

RSFLines :: { ([ReferenceExpl]) }
    : RSFLine RSFLines  { $1 : $2 }
    | RSFOther RSFLines { $2 }
    | RSFLine           { [$1] }
    | RSFOther          { [] }

RSFLine :: { ReferenceExpl }
    : REFvariable var var { (
          ("", Element{elName=unBrk2 $2,elType="MethodDecl",value=""}),
          (Element{elName=unBrk2 $3,elType="FieldDecl",value=""})) }--$2 uses field $3
    | REFvariable var wrappee var var { (
          (unBrk $4, Element{elName=$2,elType="MethodDecl",value=""}),
          (Element{elName=unBrk2 $5,elType="FieldDecl",value=""})) }
    | REFvariable var var wrappee var { (
          ("", Element{elName=unBrk2 $2,elType="MethodDecl",value=""}),
          (Element{elName=$3,elType="FieldDecl",value=""})) }
    | REFvariable var wrappee var var wrappee var { (
          (unBrk $4, Element{elName=unBrk2 $2,elType="MethodDecl",value=""}),
          (Element{elName=unBrk2 $5,elType="FieldDecl",value=""})) }
    | REFfunction var var { (
          ("", Element{elName=unBrk2 $2,elType="MethodDecl",value=""}),
          (Element{elName=unBrk2 $3,elType="MethodDecl",value=""})) }--$2 uses method $3
    | REFfunction var wrappee var var { (
          (unBrk $4, Element{elName=$2,elType="MethodDecl",value=""}),
          (Element{elName=unBrk2 $5,elType="MethodDecl",value=""})) }
    | REFfunction var var wrappee var { (
          ("", Element{elName=unBrk2 $2,elType="MethodDecl",value=""}),
          (Element{elName=$3,elType="MethodDecl",value=""})) }
    | REFfunction var wrappee var var wrappee var { (
          (unBrk $4, Element{elName=$2,elType="MethodDecl",value=""}),
          (Element{elName=$5,elType="MethodDecl",value=""})) }
-- The following reference information is ignored
--    | refClass var var { (
--          ("", Element{elName=unBrk2 $2,elType="ClassDeclaration",value=""}),
--          (Element{elName=unBrk2 $3,elType="ClassDeclaration",value=""})) } --$2 uses class $3
--    | refClass var wrappee var var { (
--          (unBrk $4, Element{elName=$2,elType="ClassDeclaration",value=""}),
--          (Element{elName=unBrk2 $5,elType="ClassDeclaration",value=""})) }
--    | refClass var var wrappee var { (
--          ("", Element{elName=unBrk2 $2,elType="ClassDeclaration",value=""}),
--          (Element{elName=$3,elType="ClassDeclaration",value=""})) }
--    | refClass var wrappee var var wrappee var { (
--          (unBrk $4, Element{elName=$2,elType="ClassDeclaration",value=""}),
--          (Element{elName=$5,elType="ClassDeclaration",value=""})) }
    | inheritClass var var { (
          ("", Element{elName=unBrk2 $2,elType="ClassDeclaration",value=""}),
          (Element{elName=unBrk2 $3,elType="ClassDeclaration",value=""})) } --$2 inherits from $3
    | inheritClass var wrappee var var { (
          (unBrk $4, Element{elName=$2,elType="ClassDeclaration",value=""}),
          (Element{elName=unBrk2 $5,elType="ClassDeclaration",value=""})) }
    | inheritClass var var wrappee var { (
          ("", Element{elName=unBrk2 $2,elType="ClassDeclaration",value=""}),
          (Element{elName=$3,elType="ClassDeclaration",value=""})) }
    | inheritClass var wrappee var var wrappee var { (
          (unBrk $4, Element{elName=$2,elType="ClassDeclaration",value=""}),
          (Element{elName=$5,elType="ClassDeclaration",value=""})) }
    | implements var var { (
          ("", Element{elName=unBrk2 $2,elType="ClassDeclaration",value=""}),
          (Element{elName=unBrk2 $3,elType="ClassDeclaration",value=""})) } --$2 inherits from $3
    | implements var wrappee var var { (
          (unBrk $4, Element{elName=$2,elType="ClassDeclaration",value=""}),
          (Element{elName=unBrk2 $5,elType="ClassDeclaration",value=""})) }
    | implements var var wrappee var { (
          ("", Element{elName=unBrk2 $2,elType="ClassDeclaration",value=""}),
          (Element{elName=$3,elType="ClassDeclaration",value=""})) }
    | implements var wrappee var var wrappee var { (
          (unBrk $4, Element{elName=$2,elType="ClassDeclaration",value=""}),
          (Element{elName=$5,elType="ClassDeclaration",value=""})) }

RSFOther :: { () }
    : MEMBER var var var { () } --Element {elName=unBrk2 $4,elType=$2,value=""}
    | MEMBER var var wrappee var var { () } 
    | MEMBER var var var wrappee var { () } 
    | MEMBER var var wrappee var var wrappee var { () } 
    | CONTAINEDIN var var { () } --parent ($3) and type ($1) of recent element
    | CONTAINEDIN var wrappee var var { () } 
    | CONTAINEDIN var var wrappee var { () } 
    | CONTAINEDIN var wrappee var var wrappee var { () } 
    | COMPOUND var var var { () } --Element {elName=unBrk2 $4,elType=$2,value=""}
    | COMPOUND var var wrappee var var { () } 
    | COMPOUND var var var wrappee var { () } 
    | COMPOUND var var wrappee var var wrappee var { () } 
    | refClass var var { () }
    | refClass var var wrappee var { () }
    | refClass var wrappee var var { () }
    | refClass var wrappee var var wrappee var { () }
    | BASEDON var var var var { () } --no idea what that is
    | LOCATEDAT var var var { () } --location ($4) of element in file ($3)
    | REFERSTO var var { () } --$2 refers to $3 (momentarily excluded from analysis)
    | refFile var var { () } --file $2 refers to file $3

{

data Token
      = TokenMember
      | TokenContainedIn
      | TokenCompound
      | TokenBasedOn
      | TokenLocatedAt
      | TokenRefersTo
      | TokenRefVar
      | TokenRefFunction
      | TokenRefClass
      | TokenRefFile
      | TokenInheritClass
      | TokenImplements
      | TokenWrappee
      | TokenElem String

instance Show Token where
    show TokenMember = "MEMBER"
    show TokenContainedIn = "CONTAINEDIN"
    show TokenCompound = "COMPOUND"
    show TokenBasedOn = "BASEDON"
    show TokenLocatedAt = "LOCATEDAT"
    show TokenRefersTo = "REFERSTO"
    show TokenRefVar = "REFvariable"
    show TokenRefFunction = "REFfunction"
    show TokenRefClass = "refClass"
    show TokenRefFile = "refFile"
    show TokenInheritClass = "inheritClass"
    show TokenImplements = "implements"
    show TokenWrappee = "__wrappee__"
    show (TokenElem str) = str

parseError :: [Token] -> a
parseError [] = error "Parse error"
parseError (x:xs) = error ("Parse error before : \"" 
    ++ (foldr (++) "" (map (' ':) (map show (take 12 (x:xs))))) 
    ++ "\"")
    
unBrk :: String -> String
unBrk [] = error "unBrk: lost brackets"
unBrk "()" = ""
unBrk (x:xs) = x:(unBrk xs)

unBrk2 :: String -> String -- only to discern from unBrk
unBrk2 [] = "" --no error here
unBrk2 "()" = ""
unBrk2 (x:xs) = x:(unBrk2 xs)

-- lexer
-- note that '"' may not be part of any reserved word
lexer :: String -> [Token]
lexer [] = []
lexer (c:cs) 
      | isSpace c = lexer cs
      | c=='"' = lexVarQuote cs
      | otherwise = lexVar (c:cs)


lexVar cs =
   case span (\x -> not(isSpace x)) cs of
      ("MEMBER",rest) -> TokenMember : lexer rest
      ("CONTAINEDIN",rest) -> TokenContainedIn : lexer rest
      ("COMPOUND",rest) -> TokenCompound : lexer rest
      ("BASEDON",rest)  -> TokenBasedOn : lexer rest
      ("LOCATEDAT",rest) -> TokenLocatedAt : lexer rest
      ("REFERSTO",rest) -> TokenRefersTo : lexer rest
      ("REFvariable",rest) -> TokenRefVar : lexer rest
      ("REFfunction",rest) -> TokenRefFunction : lexer rest
      ("refClass",rest) -> TokenRefClass : lexer rest
      ("refFile",rest) -> TokenRefFile : lexer rest
      ("inheritClass",rest) -> TokenInheritClass : lexer rest
      ("implements",rest) -> TokenInheritClass : lexer rest
      (var,rest)   -> if (isInfixOf "__wrappee__" var) 
          then let tupel = (splitAtWrappee var) 
              in ((TokenElem (fst tupel))
                 :(TokenWrappee
                    :((TokenElem (snd tupel))
                       : lexer rest))) 
          else (TokenElem var) : lexer rest

lexVarQuote cs = 
   case span (\x -> x/='"') cs of 
      (var,rest)   -> (TokenElem ('"' : (var++"\""))) : lexer (tail rest)

-- splits a string at "__wrappee__", resulting in two strings,
-- the prefix before and the suffix after.
-- The given string must contain "__wrappee__"
splitAtWrappee :: String -> (String,String)
splitAtWrappee [] = error "splitAtWrappee: given string did not contain __wrappee__"
splitAtWrappee (s:ss) | isPrefixOf "__wrappee__" (s:ss) = ("", (drop 10 ss))
                      | otherwise = let tupel = splitAtWrappee ss in
                                    (s:(fst tupel), snd tupel)


-- discards comments
removeComments :: String -> String
removeComments [] = ""
removeComments ('#':xs) = discardLine xs
    where discardLine :: String -> String --until EOL
          discardLine [] = ""
          discardLine (x:xs) | isPrint x = discardLine xs
                             | otherwise = removeComments xs
removeComments (x:xs) = x : (removeComments xs)

main = getContents >>= print . parseRSF . lexer . removeComments 

}

--happy example.y 
--happy example.y -i 



