{
--
-- Parser for extracting element names with corresponding features 
-- out of *.features.rsf files
--
-- Author: Wolfgang Scholz, 2010-06
--
module RSFextractFeatures where
import Char
import FeatureStructure


}

%name parseFeatures
-- [Token] -> ([String], [(String, Element)])
-- Result will be a tupel of lists:
-- 1) a list of feature names in the order of composition (first one first)
-- 2) a list of Elements and the Features they reside in
%tokentype { Token }
%error { parseError }

%token
    Introduces      { TokenIntroduces }
    var             { TokenElem $$ }

--%left '('
--%left not
%%

RSFLines :: { ([String], [(String, Element)]) }
    : RSFLine RSFLines  { if (head(fst $2) == fst $1) then 
                             (fst $2, $1:(snd $2)) else 
                             ((fst $1):(fst $2), $1:(snd $2)) }
    | RSFLine           { ([fst $1], [$1]) }

RSFLine :: { (String, Element) }
    : Introduces var var var { ($2, Element{ 
          elName=rmBrk $3,
          elType=(
              case $4 of --some types are not differentiated, but must be same as in reference file
                  "ConstructorDecl" -> "MethodDecl"
                  "Func"            -> "MethodDecl"
                  "StmtYL"          -> "FieldDecl"
                  _                 -> $4
              ),
          value="" --empty at this present state
          }) } 

{

data Token
      = TokenIntroduces
      | TokenElem String

instance Show Token where
    show TokenIntroduces = "Introduces"
    show (TokenElem str) = str

parseError :: [Token] -> a
parseError [] = error "Parse error"
parseError (x:xs) = error ("Parse error before : \"" 
    ++ (foldr (++) "" (map (' ':) (map show (take 12 (x:xs))))) 
    ++ "\"")

-- remove brackets
rmBrk :: String -> String
rmBrk [] = ""
rmBrk "()" = ""
rmBrk (x:xs) = x:(rmBrk xs)

-- lexer
-- note that '"' may not be part of any reserved word
lexer :: String -> [Token]
lexer [] = []
lexer (c:cs) 
      | isSpace c = lexer cs
      | otherwise = lexVar (c:cs)

lexVar cs =
   case span (\x -> not(isSpace x)) cs of --(isAlpha x) || (isDigit x)
      ("Introduces",rest) -> TokenIntroduces : lexer rest
      (var,rest)   -> (TokenElem var) : lexer rest

-- discards comments
removeComments :: String -> String
removeComments [] = ""
removeComments ('#':xs) = discardLine xs
    where discardLine :: String -> String --until EOL
          discardLine [] = ""
          discardLine (x:xs) | isPrint x = discardLine xs
                             | otherwise = removeComments xs
removeComments (x:xs) = x : (removeComments xs)

main = getContents >>= print . parseFeatures . lexer . removeComments 

}

--happy example.y 
--happy example.y -i 



